# -*- coding: utf-8 -*-

"""
Collection of helper functions being used in lbsntransform package.
"""


import datetime as dt
import importlib.util
import json
import logging
# due to different protocol buffers implementations on Unix, MacOS and Windows
# import types based on OS
import platform
import re
import string
from datetime import timezone
from json import JSONDecodeError, JSONDecoder
from pathlib import Path
from typing import List, Optional, Set, Union

import emoji
import regex
from google.protobuf.timestamp_pb2 import Timestamp
from lbsnstructure import lbsnstructure_pb2 as lbsn
from lbsntransform.output.shared_structure import Coordinates
from shapely import geos, wkt
from shapely.geometry import Point, Polygon

PLATFORM_SYS = platform.system()
if PLATFORM_SYS == 'Linux':
    from google.protobuf.pyext._message import \
        RepeatedCompositeContainer  # pylint: disable=no-name-in-module
    from google.protobuf.pyext._message import \
        ScalarMapContainer  # pylint: disable=no-name-in-module
else:
    from google.protobuf.internal.containers import \
        RepeatedCompositeFieldContainer  # pylint: disable=no-name-in-module
    from google.protobuf.internal.containers import \
        ScalarMapContainer  # pylint: disable=no-name-in-module

NLTK_AVAIL = True
try:
    # check if nltk is installed
    import nltk
except ImportError:
    NLTK_AVAIL = False

if NLTK_AVAIL:
    try:
        # check if stopwords corpus is available
        from nltk.corpus import stopwords
    except LookupError:
        print(
            'Please use '
            '`python -c \'import nltk;nltk.download("stopwords")\'` '
            'to install stopwords resource globally. Continuing without '
            'nltk stopwords filter..')
        NLTK_AVAIL = False
# pylint: disable=no-member


class HelperFunctions():
    """Collection of helper functions being used in lbsntransform package"""

    # Null Geometry String (4326)
    # for improving performance in PostGIS Upserts
    NULL_GEOM_HEX = '0101000020E610000000000000000000000000000000000000'

    @staticmethod
    def value_count(value_x: str):
        """Turn none values into 0, otherwise return value"""
        if value_x is None:
            return 0
        if isinstance(value_x, int):
            return value_x
        return int(value_x) if value_x.isdigit() else 0

    @staticmethod
    def remove_prefix(text_str: str, prefix: str):
        """Remove prefix from string"""
        if text_str.startswith(prefix):
            return text_str[len(prefix):]
        return text_str

    @staticmethod
    def concat_values_str(
            sql_escaped_values_list: List[str]) -> str:
        """Concat List of sql escaped values with comma separated
        """
        values_str = ','.join(sql_escaped_values_list)
        return values_str

    @staticmethod
    def sanitize_string(str_text: str):
        """Sanitize text strings for postgres sql compatibility

        * remove any NUL (0x00) characters
        """
        return str_text.replace('\x00', '')

    @staticmethod
    def format_base_repr(base):
        """Return formatted string of base"""
        return (
            f'{base.NAME.base}\nFacet: {base.NAME.facet}, '
            f'Key: {base.get_key_value()}, '
            f'Metrics: \n'
            f'{[":".join([k, str(len(v))]) for k, v in base.metrics.items()]}')

    @staticmethod
    def remove_hyperlinks(text_s):
        """Remove any hyperlinks from string (regex)

        Note:
        - anything between <a>xxx</a> will be kept
        """
        pattern = r'<(a|/a).*?>'
        result = re.sub(pattern, "", text_s)
        return result

    @staticmethod
    def get_all_post_terms(
            record: Optional[lbsn.Post] = None) -> Set[str]:
        """Returns all post terms combined in single set"""
        body_terms = HelperFunctions.select_terms(
            record.post_body)
        title_terms = HelperFunctions.select_terms(
            record.post_title)
        tag_terms = HelperFunctions.filter_terms(
            record.hashtags)
        all_post_terms = set.union(
            body_terms, title_terms, tag_terms)
        return all_post_terms

    @staticmethod
    def get_all_post_emoji(
            post_body: str) -> Set[str]:
        """Returns all post emoji combined in single set"""
        return set(HelperFunctions.extract_emoji(post_body))

    @staticmethod
    def select_terms(
            text_s: str, selection_list: List[str] = None) -> Set[str]:
        """Extract list of words from sentence and return filtered version
        """
        # first remove hyperlinks
        text_s = HelperFunctions.remove_hyperlinks(text_s)
        # remove problematic characters from string
        text_s = HelperFunctions.sanitize_string(text_s)
        # remove punctuation
        text_s = text_s.translate(
            str.maketrans('', '', string.punctuation))
        # split string by space character into list
        querywords = text_s.split()
        resultwords = HelperFunctions.filter_terms(
            querywords, selection_list)
        return resultwords

    @staticmethod
    def filter_terms(
            terms: List[str],
            selection_list: List[str] = None) -> Set[str]:
        """Filter a list of terms

        * based on a provided positive(negative) filter list of terms,
        * based on length (minimum 2 characters),
        * based on type (no plain numbers)
        """
        # turn lowercase
        querywords = [word.lower() for word in terms]
        # filter based on
        # - stoplist/selectionlist
        # - length (3+ character)
        # - type: no numbers
        resultwords = {
            word for word in querywords if (
                selection_list is None or word in selection_list)
            and len(word) > 2
            and HelperFunctions.nltk_stopword_filter(word)
            and not word.isdigit()}
        return resultwords

    @staticmethod
    def nltk_stopword_filter(term: str) -> bool:
        """Filter term against nltk stopwords (english)"""
        if NLTK_AVAIL:
            if term in stopwords.words('english'):
                return False
        return True

    @staticmethod
    def reduce_ewkt_to_wkt(geom_ewkt: str) -> str:
        """Hack to reduce extended WKT (eWKT) to WKT"""
        geom_wkt = geom_ewkt.replace("SRID=4326;", "")
        return geom_wkt

    @staticmethod
    def get_geom_from_ewkt(
            geom_ewkt: str) -> Union[Point, Polygon]:
        """Convert EWKT representation (without srid) to shapely geometry

        Note: either Point or Polygon
        """
        geom_wkt = HelperFunctions.reduce_ewkt_to_wkt(geom_ewkt)
        shply_geom = wkt.loads(geom_wkt)
        return shply_geom

    @staticmethod
    def get_coordinates_from_ewkt(
            geom: str) -> Coordinates:
        """Convert EWKT representation (with srid) to geometry

        Note:
        Shapely has no support for handling SRID (projection). The
        approach used here is a shortcut. This should be replaced
        by proper EWKT handling using a package, e.g.
        django.contrib.gis.geos or django.contrib.gis.geometry, see:
        https://docs.huihoo.com/django/1.11/ref/contrib/gis/geos.html
        """
        if not geom:
            return Coordinates()
        geom = HelperFunctions.reduce_ewkt_to_wkt(geom)
        shply_geom = HelperFunctions.get_geom_from_ewkt(geom)
        if not shply_geom.geom_type == "Point":
            raise ValueError(
                f"Expected geometry of type Point, "
                f"but found {shply_geom.geom_type}")
        coordinates = Coordinates(
            lng=shply_geom.x, lat=shply_geom.y)  # pylint: disable=maybe-no-member
        return coordinates

    @staticmethod
    def extract_hashtags_from_string(
        text_str: str) -> Set[str]:
        """Extract hashtags with leading hash-character (#) from string

        - removes # from hashtags
        - removes duplicates
        - removes special chars (emoji etc.) from hashtags, e.g.:
            - input: "#germany🇩🇪"
            - output: [germany]
        """
        hashtag_pattern = re.compile('(?i)(?<=\#)\w+')
        hashtag_list = hashtag_pattern.findall(text_str)
        return set(hashtag_list)

    @staticmethod
    def extract_atmentions_from_string(
        text_str: str) -> Set[str]:
        """Extract @-mentions with leading hash-character (@) from string

        - removes @ from mentions
        - removes duplicates
        - removes special chars (emoji etc.) from mentions, e.g.:
            - input: "@userxyz🇩🇪"
            - output: [userxyz]
        """
        mention_pattern = re.compile('(?i)(?<=\@)\w+')
        mention_list = mention_pattern.findall(text_str)
        return set(mention_list)

    @staticmethod
    def json_read_wrapper(gen):
        """Wraps json iterator and catches any error"""
        while True:
            try:
                yield next(gen)
            except StopIteration:
                # no further items produced by the iterator
                raise
            except json.decoder.JSONDecodeError:
                HelperFunctions._log_json_decodeerror(gen)
            except Exception as e:
                HelperFunctions._log_unhandled_exception(e)

    @staticmethod
    def json_load_wrapper(gen, single: bool = None):
        """Wraps json load(s) and catches any error"""
        if single is None:
            single = False
        try:
            if single:
                record = json.loads(gen)
                return record
            records = json.load(gen)
            return records
        except json.decoder.JSONDecodeError:
            HelperFunctions._log_json_decodeerror(gen)
        except Exception as exc_general:
            HelperFunctions._log_unhandled_exception(exc_general)

    @staticmethod
    def _log_json_decodeerror(record_str: str):
        logging.getLogger('__main__').warning(
            f"\nJSONDecodeError: skipping entry\n{record_str}\n\n")

    @staticmethod
    def _log_unhandled_exception(e: str):
        logging.getLogger('__main__').warning(
            f"\nUnhandled exception: \n{e}\n ..skipping entry\n")

    @staticmethod
    def report_stats(input_cnt, current_cnt, lbsn_records=None):
        """Format string for reporting stats."""
        report_stats = (f'{input_cnt} '
                        f'input records read (up to '
                        f'{current_cnt}). '
                        f'{HelperFunctions.get_count_stats(lbsn_records)}')
        return report_stats

    @staticmethod
    def get_count_stats(lbsn_records=None):
        """Format string for reporting count stats."""
        if lbsn_records is None:
            return
        report_stats = (
            f'Count per type: '
            f'{lbsn_records.get_type_counts()}'
            f'records.')
        return report_stats

    @staticmethod
    def get_str_formatted_today():
        """Returns date as string (YYYY-mm-dd)"""
        today = dt.date.today()
        today_str = today.strftime("%Y-%m-%d")
        return today_str

    @staticmethod
    def set_logger():
        """ Set logging handler manually,
        so we can also print to console while logging to file
        """

        logging.basicConfig(
            handlers=[logging.FileHandler(
                'log.log', 'w', 'utf-8')],
            format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
            datefmt='%H:%M:%S',
            level=logging.DEBUG)
        log = logging.getLogger(__name__)

        # Get Stream handler
        logging.getLogger().addHandler(logging.StreamHandler())
        return log

    @staticmethod
    def geoacc_within_threshold(post_geoaccuracy, min_geoaccuracy):
        """Checks if geoaccuracy is within or below threshhold defined"""
        if min_geoaccuracy == lbsn.Post.LATLNG:
            allowed_geoaccuracies = [lbsn.Post.LATLNG]
        elif min_geoaccuracy == lbsn.Post.PLACE:
            allowed_geoaccuracies = [lbsn.Post.LATLNG, lbsn.Post.PLACE]
        elif min_geoaccuracy == lbsn.Post.CITY:
            allowed_geoaccuracies = [lbsn.Post.LATLNG, lbsn.Post.PLACE,
                                     lbsn.Post.CITY]
        else:
            return True
        # check post geoaccuracy
        return bool(post_geoaccuracy in allowed_geoaccuracies)

    @staticmethod
    def get_version():
        """Gets the program version number from version file in root"""
        with open('VERSION') as version_file:
            version_var = version_file.read().strip()
        return version_var

    @staticmethod
    def log_main_debug(debug_text):
        """Issues a main debug log in case it is
        needed for static functions.
        """
        logging.getLogger('__main__').debug(debug_text)

    @staticmethod
    def null_notice(x_value: int) -> str:
        """Reporting: Suppresses null notice (for Null island)
        if value is zero.
        """
        return f'(Null Island: {x_value})' if x_value > 0 else ''

    @staticmethod
    def utc_to_local(utc_dt):
        """Convert utc to local time"""
        return utc_dt.replace(tzinfo=timezone.utc).astimezone(tz=None)

    @staticmethod
    def cleanhtml(raw_html: str) -> str:
        """Remove any html tags from string"""
        cleanr = re.compile('<.*?>')
        cleantext = re.sub(cleanr, '', raw_html)
        return cleantext

    @staticmethod
    def extract_emoji_fast(str_text: str):
        """Extract list of emoji from string

        Note:
            cannot detect grapheme clusters or flags"""
        # str = str.decode('utf-32').encode('utf-32', 'surrogatepass')
        # return list(c for c in str if c in emoji.UNICODE_EMOJI)
        return set(c for c in str_text if c in emoji.UNICODE_EMOJI)

    @staticmethod
    def extract_emoji(string_with_emoji: str) -> Set[str]:
        """Extract emoji and flags using regex package

        This is a new version to extract emoji (see old method:
        _extract_emoji_old). Code source:
        https://stackoverflow.com/a/49242754/4556479
        This method supports extracting grapheme clusters,
        emoji constructed of multiple emoji (the "perceived
        pictograms")

        Compare:
        A: _extract_emoji_old:
        Total emoji count for the 100 most
        used emoji in selected area: 27722.
        Total distinct emoji (DEC): 918
        B: _extract_emoji:
        Total emoji count for the 100 most
        used emoji in selected area: 25793.
        Total distinct emoji (DEC): 1349

        Original Code @ sheldonzy:
        https://stackoverflow.com/a/49242754/4556479
        """
        emoji_set = set()
        # use \X (eXtended grapheme cluster) regular expression:
        data = regex.findall(r'\X', string_with_emoji)
        for grapheme in data:
            if any(char in emoji.UNICODE_EMOJI["en"] for char in grapheme):
                emoji_set.add(grapheme)
        return emoji_set

    @staticmethod
    def get_rectangle_bounds(points):
        """Get rectangle boundary from list of points"""
        lats = []
        lngs = []
        for point in points:
            lngs.append(point[0])
            lats.append(point[1])
        lim_y_min = min(lats)
        lim_y_max = max(lats)
        lim_x_min = min(lngs)
        lim_x_max = max(lngs)
        return lim_y_min, lim_y_max, lim_x_min, lim_x_max

    @staticmethod
    def new_lbsn_record_with_id(record, id, origin):
        """Initialize new lbsn record with composite ID"""
        c_key = lbsn.CompositeKey()
        c_key.origin.CopyFrom(origin)
        c_key.id = id
        record.pkey.CopyFrom(c_key)
        return record

    @staticmethod
    def new_lbsn_relation_with_id(
            lbsn_relationship, relation_to_id,
            relation_from_id, relation_origin):
        """Initialize new lbsn relationship with 2 composite IDs
        for one origin
        """
        c_key_to = lbsn.CompositeKey()
        c_key_to.origin.CopyFrom(relation_origin)
        c_key_to.id = relation_to_id
        c_key_from = lbsn.CompositeKey()
        c_key_from.origin.CopyFrom(relation_origin)
        c_key_from.id = relation_from_id
        r_key = lbsn.RelationshipKey()
        r_key.relation_to.CopyFrom(c_key_to)
        r_key.relation_from.CopyFrom(c_key_from)
        lbsn_relationship.pkey.CopyFrom(r_key)
        return lbsn_relationship

    @staticmethod
    def is_post_reaction(json_string):
        """Determine if post is post reaction

        The retweeted field will return true if a tweet _got_ retweeted
        To detect if a tweet is a retweet of other tweet,
        check the retweeted_status field
        """
        return bool('quoted_status' in json_string
                    or 'retweeted_status' in json_string
                    or json_string.get('in_reply_to_status_id_str'))

    @staticmethod
    def assign_media_post_type(json_media_string):
        """Media type assignment based on Twitter json"""
        # if post, get type of first entity
        type_string = json_media_string[0].get('type')
        # type is either photo, video, or animated_gif
        # https://developer.twitter.com/en/docs/tweets/data-dictionary/overview/extended-entities-object.html
        if type_string:
            if type_string == "photo":
                post_type = lbsn.Post.IMAGE
            elif type_string in ("video", "animated_gif"):
                post_type = lbsn.Post.VIDEO

        else:
            post_type = lbsn.Post.OTHER
            logging.getLogger('__main__').debug(f'Other lbsn.Post type detected: '
                                                f'{json_media_string}')
        return post_type

    @staticmethod
    def json_date_string_to_proto(json_date_string: str):
        """Parse String -Date Format found in Twitter json"""
        date_time_record = dt.datetime.strptime(
            json_date_string, '%a %b %d %H:%M:%S +0000 %Y')
        protobuf_timestamp_record = Timestamp()
        # Convert to ProtoBuf Timestamp Recommendation
        protobuf_timestamp_record.FromDatetime(date_time_record)
        return protobuf_timestamp_record

    @staticmethod
    def json_date_timestamp_to_proto(json_date_timestamp):
        """Parse String -Timestamp Format found in Twitter json"""
        date_time_record = dt.datetime.fromtimestamp(
            json_date_timestamp)
        protobuf_timestamp_record = Timestamp()
        # Convert to ProtoBuf Timestamp Recommendation
        protobuf_timestamp_record.FromDatetime(date_time_record)
        return protobuf_timestamp_record

    @staticmethod
    def parse_csv_datestring_to_protobuf(csv_datestring, t_format=None):
        """Parse String -Timestamp Format found in Flickr csv

        e.g. 2012-02-16 09:56:37.0
        """
        if t_format is None:
            t_format = '%m/%d/%Y %H:%M:%S'
        try:
            date_time_record = dt.datetime.strptime(
                csv_datestring, t_format)
        except ValueError:
            return None
        protobuf_timestamp_record = Timestamp()
        # Convert to ProtoBuf Timestamp Recommendation
        protobuf_timestamp_record.FromDatetime(date_time_record)
        return protobuf_timestamp_record

    @staticmethod
    def parse_timestamp_string_to_protobuf(timestamp_string):
        """Converts timestamp string to protobuf object"""
        # Parse from RFC 3339 date string to Timestamp.
        time_date = dt.datetime.fromtimestamp(int(timestamp_string))
        protobuf_timestamp_record = Timestamp()
        protobuf_timestamp_record.FromDatetime(time_date)
        return protobuf_timestamp_record

    @staticmethod
    def get_mentioned_users(userMentions_jsonString, origin):
        """Return list of mentioned users from json"""
        mentioned_users_list = []
        for user_mention in userMentions_jsonString:  # iterate over the list
            ref_user_record = \
                HelperFunctions.new_lbsn_record_with_id(
                    lbsn.User(), user_mention.get('id_str'), origin)
            ref_user_record.user_fullname = \
                user_mention.get('name')  # Needs to be saved
            ref_user_record.user_name = user_mention.get('screen_name')
            mentioned_users_list.append(ref_user_record)
        return mentioned_users_list

    @staticmethod
    def substitute_referenced_user(main_post, origin, log):
        """Look for mentioned userRecords"""
        ref_user_pkey = None
        user_mentions_json = main_post.get('entities').get('user_mentions')
        if user_mentions_json:
            ref_user_records = \
                HelperFunctions.get_mentioned_users(user_mentions_json,
                                                    origin)
            # if it is a retweet, and the status contains 'RT @',
            # and the mentioned UserID is also in the status,
            # we can almost be completely certain that it is the userid who
            # posted the original tweet that was retweeted
            if ref_user_records \
               and ref_user_records[0].user_name.lower() in \
               main_post.get('text').lower() \
               and main_post.get('text').startswith(f'RT @'):
                ref_user_pkey = ref_user_records[0].pkey
            if ref_user_pkey is None:
                log.warning(
                    f'No lbsn.User record found for referenced post in: '
                    f'{main_post}')
                input("Press Enter to continue... "
                      "(post will be saved without userid)")
        return ref_user_pkey

    @staticmethod
    def null_check(record_attr):
        """Helper function to check for Null Values
        """
        if not record_attr:
            # will catch empty and None
            return None
        # This function will also remove Null bytes from string,
        # which aren't supported by Postgres
        if isinstance(record_attr, str):
            record_attr = HelperFunctions.clean_null_bytes_from_str(
                record_attr)
        return record_attr

    @staticmethod
    def null_geom_check(geom_attr):
        """Helper function to check for Null Values
        in geometry columns and replace with Null Island

        Note:
        null_geom_check is only applied to geometry columns
        with NOT NULL Constraint
        """
        if geom_attr is None or \
                (isinstance(geom_attr, str) and geom_attr == ''):
            null_island = "POINT(%s %s)" % (0, 0)
            return null_island
        return geom_attr

    @staticmethod
    def null_check_datetime(recordAttr):
        """Check if date is null or empty and replace with default value"""
        if not recordAttr:
            # will catch empty and None
            return None
        try:
            dt_attr = recordAttr.ToDatetime()
        except:
            return None
        if dt_attr == dt.datetime(1970, 1, 1, 0, 0, 0):
            return None
        return recordAttr.ToDatetime()

    @staticmethod
    def return_ewkb_from_geotext(text):
        """Generates Geometry in Well-known-Text format
        from PostGis Format (e.g. 'POINT(0 0)')
        with SRID for WGS1984 (4326)

        Note that:
        geos.WKBWriter.defaults['include_srid'] = True
        must be set (see config.py)
        """
        if text is None:
            # keep Null geometries, e.g. for geom_area columns
            return None
        geom = wkt.loads(text)
        # Set SRID to WGS1984
        geos.lgeos.GEOSSetSRID(geom._geom, 4326)
        geom = geom.wkb_hex
        return geom

    @staticmethod
    def decode_stacked(document, pos=0, decoder=JSONDecoder()):
        """Decode stacked json"""
        not_whitespace = re.compile(r'[^\s]')
        while True:
            match = not_whitespace.search(document, pos)
            if not match:
                return
            pos = match.start()

            try:
                obj, pos = decoder.raw_decode(document, pos)
            except JSONDecodeError:
                raise
            yield obj

    @staticmethod
    def clean_null_bytes_from_str(text_str: str):
        """Remove null bytes from string for pg compatibility"""
        str_without_null_byte = text_str.replace('\x00', '')
        return str_without_null_byte

    @staticmethod
    def turn_lower(text_str):
        """Returns lower but keeps none values"""
        if text_str:
            return text_str.lower()
        return text_str

    @staticmethod
    def empty_list(list_str):
        """Returns lower but keeps none values"""
        if list_str:
            if len(list_str) == 0:
                return None
            else:
                return list_str
        return None

    @staticmethod
    def is_composite_field_container(in_obj):
        """Checks whether in_obj is of type RepeatedCompositeFieldContainer"""
        if PLATFORM_SYS == 'Linux':
            if isinstance(
                    in_obj, (RepeatedCompositeContainer, ScalarMapContainer)):
                return True
            return False
        if isinstance(
                in_obj, (RepeatedCompositeFieldContainer, ScalarMapContainer)):
            return True
        return False

    @staticmethod
    def map_to_dict(proto_map):
        """Converts protobuf field map (ScalarMapContainer)
        to Dictionary"""
        if proto_map:
            mapped_dict = dict(zip(proto_map.keys(), proto_map.values()))
            return mapped_dict
        return {}

    @staticmethod
    def _get_file_list(path: Path, ext: str = "py"):
        """Return file list in folder"""
        return [file.stem for file in path.glob(f'*.{ext}')]

    @staticmethod
    def dynamic_get_mapping_module(
            origin: int,
            mappings_path: Path = None):
        """Function to dynamically register input mappings

        Args:
            origin: The MAPPING_ID to identify the mapping module.
            path: Override default path with user defined folder.
        """
        if mappings_path is None or origin == 0:
            mappings_module_name = "lbsntransform.input.mappings"
            from lbsntransform.input.mappings.field_mapping_lbsn import \
                importer as importer
            return importer
        else:
            mapping_modules = HelperFunctions._get_file_list(
                mappings_path)
            init_file_str = "__init__"
            mappings_module_name = mappings_path.name
            for mapping_module in mapping_modules:
                if mapping_module == init_file_str:
                    continue
                spec = importlib.util.spec_from_file_location(
                    f"{mappings_path.name}.{mapping_module}",
                    mappings_path / f'{mapping_module}.py')
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                if hasattr(module, 'MAPPING_ID') and module.MAPPING_ID == origin:
                    if hasattr(module, 'importer'):
                        importer = module.importer
                        return importer
                    raise ImportError("importer missing in {module}")
        raise ValueError(
            f'{origin} not found in {mappings_module_name}. '
            f'Input type not supported.')

    @staticmethod
    def load_module(package: str, name: str):
        name = f"{package}.{name}"
        __import__(name, fromlist=[''])

    @staticmethod
    def load_importer_mapping_module(
            origin: int, mappings_path: Path = None):
        """ Switch import module based on origin input
            1 - Instagram, 2 - Flickr, 3 - Twitter, 4 - Facebook
        """
        importer = HelperFunctions.dynamic_get_mapping_module(
            origin=origin, mappings_path=mappings_path)
        return importer

    @staticmethod
    def dict_type_switcher(desc_name):
        """ Create protoBuf messages by name"""
        dict_switcher = {
            lbsn.Country().DESCRIPTOR.name: lbsn.Country(),
            lbsn.City().DESCRIPTOR.name: lbsn.City(),
            lbsn.Place().DESCRIPTOR.name: lbsn.Place(),
            lbsn.User().DESCRIPTOR.name: lbsn.User(),
            lbsn.UserGroup().DESCRIPTOR.name:  lbsn.UserGroup(),
            lbsn.Post().DESCRIPTOR.name: lbsn.Post(),
            lbsn.PostReaction().DESCRIPTOR.name: lbsn.PostReaction(),
            lbsn.Relationship().DESCRIPTOR.name: lbsn.Relationship()
        }
        return dict_switcher.get(desc_name)

    @staticmethod
    def check_notice_empty_post_guid(post_guid):
        """Check if post_guid empty and if, raise warning"""
        if not post_guid:
            logging.getLogger('__main__').warning(
                f'No PostGuid\n\n'
                f'{post_guid}')
            return False
        return True

    @staticmethod
    def get_skipped_report(import_mapper):
        """Get count report of records skipped due to low geoaccuracy
        or ignore list"""
        skipped_geo = None
        skipped_ignore = None
        # check if methods habe been implemented in import mapper module
        try:
            skipped_geo_count = import_mapper.get_skipped_geoaccuracy()
        except AttributeError:
            skipped_geo_count = 0
        try:
            skipped_ignorelist_count = import_mapper.get_skipped_ignorelist()
        except AttributeError:
            skipped_ignorelist_count = 0
        # compile report texts
        if skipped_geo_count > 0:
            skipped_geo = (f'Skipped '
                           f'{skipped_geo_count} '
                           f'due to low geoaccuracy.')
        if skipped_ignorelist_count > 0:
            skipped_ignore = (f'Skipped '
                              f'{skipped_ignorelist_count} '
                              f'due to ignore list.')
        if skipped_geo is None and skipped_ignore is None:
            return ''
        else:
            report_str = ' '.join([skipped_geo, skipped_ignore])
            return report_str



# -*- coding: utf-8 -*-

"""Hll bases for spatial facet
"""

from typing import Union

from lbsnstructure import lbsnstructure_pb2 as lbsn
from lbsntransform.output.hll import hll_bases as hll
from lbsntransform.tools.helper_functions import HelperFunctions as HF

from lbsntransform.output.hll.hll_functions import HLLFunctions as HLF

FACET = 'spatial'


class LatLngBase(hll.HllBase):
    """Extends Base Class"""
    # base reference, eg.
    # e.g.: facet: temporal, base: date
    NAME = hll.HllBaseRef(facet=FACET, base='latlng')

    def __init__(self, record: lbsn.Post = None):
        super().__init__()
        # init key and any additional attributes
        self.key['latitude'] = None
        self.key['longitude'] = None
        self.attrs['latlng_geom'] = None
        # init additional metrics
        # beyond those defined inHllBase
        self.metrics['pud_hll'] = set()
        self.metrics['utl_hll'] = set()
        if record is None:
            # init empty
            return
        if isinstance(record, lbsn.Post):
            coordinates_geom = record.post_latlng
            coordinates = HF.get_coordinates_from_ewkt(
                coordinates_geom
            )
            self.key['latitude'] = coordinates.lat
            self.key['longitude'] = coordinates.lng
            # additional (optional) attributes
            # formatted ready for sql upsert
            self.attrs['latlng_geom'] = HF.return_ewkb_from_geotext(
                coordinates_geom)
        else:
            raise ValueError(
                "Parsing of LatLngBase only supported "
                "from lbsn.Post")


class PlaceBase(hll.HllBase):
    """Extends HllBase Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='place')

    def __init__(self, record: Union[
            lbsn.Post,
            lbsn.Place] = None):
        super().__init__()
        self.key['place_guid'] = None
        self.attrs['geom_center'] = None
        self.attrs['geom_area'] = None
        self.attrs['name'] = None
        self.metrics['pud_hll'] = set()
        self.metrics['utl_hll'] = set()
        if record is None:
            return
        if isinstance(record, lbsn.Post):
            # Post can be of Geoaccuracy "Place" without any
            # actual place id assigned (e.g. Flickr Geoaccuracy level < 10)
            # in this case, concat lat:lng as primary key
            coordinates_geom = record.post_latlng
            if not record.place_pkey.id:
                coordinates = HF.get_coordinates_from_ewkt(
                    coordinates_geom
                )
                self.key['place_guid'] = HLF.hll_concat(
                    [coordinates.lat, coordinates.lng])
            else:
                self.key['place_guid'] = record.place_pkey.id
            # additional (optional) attributes
            # formatted ready for sql upsert
            self.attrs['geom_center'] = HF.return_ewkb_from_geotext(
                coordinates_geom)
            # geom_area not available from lbsn.Post
        elif isinstance(record, lbsn.Place):
            coordinates_geom = record.geom_center
            coordinates = HF.get_coordinates_from_ewkt(
                coordinates_geom
            )
            self.key['place_guid'] = record.pkey.id
            # self.key['place_guid'] = HLF.hll_concat(
            #     [coordinates.lat, coordinates.lng])
            self.attrs['geom_center'] = HF.return_ewkb_from_geotext(
                coordinates_geom)
            self.attrs['geom_area'] = HF.return_ewkb_from_geotext(
                HF.null_check(record.geom_area))
            self.attrs['name'] = record.name


class SpatialBase(hll.HllBase):
    """Intermediate spatial base class that extends HllBase
    for Place, City, Region and Country
    sharing common attributes
    """

    def __init__(self, record: Union[
            lbsn.Post,
            lbsn.Place,
            lbsn.City,
            lbsn.Country] = None):
        super().__init__()
        self.key["guid"] = None
        self.attrs['name'] = None
        self.attrs['geom_center'] = None
        self.attrs['geom_area'] = None
        self.metrics['pud_hll'] = set()
        self.metrics['utl_hll'] = set()
        self.metrics['latlng_hll'] = set()
        if record is None:
            # init empty
            return
        name = None
        geom_area = None
        if isinstance(record, lbsn.Post):
            coordinates_geom = record.post_latlng
            coordinates = HF.get_coordinates_from_ewkt(
                coordinates_geom
            )
            # use concat lat:lng as key of no place_key available
            # this should later implement assignemnt based on area
            # intersection
            self.key["guid"] = HLF.hll_concat(
                [coordinates.lat, coordinates.lng])
        elif isinstance(record, (lbsn.Place, lbsn.City, lbsn.Country)):
            name = HF.null_check(record.name)
            coordinates_geom = record.geom_center
            geom_area = record.geom_area
            # use key from place, city or country record
            self.key["guid"] = HLF.hll_concat_origin_guid(record)

        self.attrs['name'] = name
        self.attrs['geom_center'] = HF.return_ewkb_from_geotext(
            coordinates_geom)
        self.attrs['geom_area'] = HF.return_ewkb_from_geotext(
            geom_area)


class CityBase(SpatialBase):
    """Extends Spatial Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='city')

    def __init__(self, record: lbsn.Post = None):
        super().__init__(record)
        # rename key accordingly
        self.key["city_guid"] = self.key.pop("guid")


class RegionBase(SpatialBase):
    """Extends Spatial Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='region')

    def __init__(self, record: lbsn.Post = None):
        super().__init__(record)
        # rename key accordingly
        self.key["region_guid"] = self.key.pop("guid")


class CountryBase(SpatialBase):
    """Extends Spatial Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='country')

    def __init__(self, record: lbsn.Post = None):
        super().__init__(record)
        # rename key accordingly
        self.key["country_guid"] = self.key.pop("guid")



# -*- coding: utf-8 -*-

"""
Collection of functions used in hll transformation
"""

import datetime as dt
from typing import Dict, Generator, List, Optional, Set, Tuple, Union

import psycopg2

from lbsnstructure import lbsnstructure_pb2 as lbsn
from lbsntransform.tools.helper_functions import HelperFunctions as HF


class HLLFunctions():
    """Collection of functions used in hll transformation"""

    @staticmethod
    def merge_hll_tuples(
        base_tuples: List[Tuple[Union[str, int], ...]],
        hll_tuples: List[Tuple[Union[str, int], ...]]) -> Generator[
            Tuple[Union[int, str], ...], None, None]:
        """Merges two lists of tuples,
        whereby only the third part of hll_tuples (the hll shard)
        is appended. The first part of base_tuples is used to collect hll
        shards per base"""
        # assign start values
        hll_baseidx_before = hll_tuples[0]
        merged_shard_batch = (hll_baseidx_before[2],)
        ix_before = hll_baseidx_before[0]
        # start from second element
        for hll_tup in hll_tuples[1:]:
            hll_shard = hll_tup[2]
            ix = hll_tup[0]
            if ix == ix_before:
                # if base is unchanged (and not the last hll_tuple),
                # concat hll shards per base into one tuple
                merged_shard_batch += (hll_shard,)
                ix_before = ix
                # continue with next shard
                continue
            # else: if base changed,
            # append hll_shards to base record
            merged_base_hll_tuple = base_tuples[ix_before] + merged_shard_batch
            merged_shard_batch = (hll_shard,)
            ix_before = ix
            # return merged record + hll tuple
            # and continue iteration
            yield merged_base_hll_tuple
        # final return of last item
        yield base_tuples[ix] + merged_shard_batch

    @staticmethod
    def concat_base_shards(
        records: List[Tuple[Union[str, int], ...]],
        hll_shards: List[Tuple[Union[str, int], ...]]) -> List[
            Tuple[Union[int, str], ...]]:
        """Concat records with list of hll_shards

        For ease of calculation, hll_shards are reduced to
        a onedimensional list. Here, shards are assigned back to
        records using record base keys
        """
        merged_records = [
            merged_tuple for merged_tuple in
            HLLFunctions.merge_hll_tuples(records, hll_shards)]
        # if this ever evaluates to false
        # something went fundamentally wrong:
        # abort processing
        assert len(records) == len(merged_records)
        return merged_records

    @staticmethod
    def make_shard_sql(values_str: str) -> str:
        """SQL to calculate HLL Shards from sets of items

        Example values_str:
        ('2011-01-01'::date, 'user_hll', '2:4399297324604'),
        """
        insert_sql = \
            f'''
            SELECT
                base_id,
                metric_id,
                hll_add_agg(
                    hll_hash_text(
                        extensions.crypt_hash(
                            s.item_value,
                            current_setting('crypt.salt'))))
            FROM (
            VALUES {values_str}
            ) s(base_id, metric_id, item_value)
            GROUP BY base_id, metric_id
            ORDER BY base_id, metric_id
            '''
        return insert_sql

    @staticmethod
    def calculate_item_shards(
        hllworker_cursor, values_str: str
    ) -> List[Tuple[int, str, str]]:
        """Calculates shards from batched hll_items
        using hll_worker connection
        """
        tsuccessful = False
        shard_sql = HLLFunctions.make_shard_sql(values_str)
        while not tsuccessful:
            try:
                hllworker_cursor.execute(shard_sql)
            except psycopg2.OperationalError:
                input(
                    "Server closed the connection unexpectedly. Check "
                    "if your postgres server has enough memory available. "
                    "Will repeat execution once after input..")
                hllworker_cursor.execute(shard_sql)
                tsuccessful = True
            else:
                tsuccessful = True
        hll_shards = hllworker_cursor.fetchall()
        return hll_shards

    @staticmethod
    def concat_base_metric_item(
            base_key: int,
            hll_items: Dict[str, set]) -> List[Tuple[int, int, str]]:
        """Concat hll item to (base_key, metric_key, item) tuple
        Note that only ids (0, 1, 2..), not the true keys are used
        for base_key and metric_key,
        following the separation of concerns principle
        """
        tuple_list = []
        for metric_ix, item_set in enumerate(hll_items.values()):
            if not item_set:
                tuple_list.append(
                    (base_key, metric_ix, None))
            else:
                for item in item_set:
                    tuple_list.append(
                        (base_key, metric_ix, item))
        return tuple_list

    @staticmethod
    def hll_concat_origin_guid(record) -> str:
        """Concat origin and pkey (guid) to string"""
        origin_guid = HLLFunctions.hll_concat(
            [record.pkey.origin.origin_id, record.pkey.id])
        return origin_guid

    @staticmethod
    def hll_concat_user(record: lbsn.Post) -> str:
        """Concat origin and user pkey (guid) to string"""
        user_hll = HLLFunctions.hll_concat(
            [record.pkey.origin.origin_id, record.user_pkey.id])
        return user_hll

    @staticmethod
    def hll_concat_date(record: lbsn.Post) -> str:
        """Merge create and publish date and concat to string"""
        date_merged = HLLFunctions.merge_dates_post(record)
        if date_merged is None:
            return
        date_formatted = date_merged.strftime('%Y-%m-%d')
        return date_formatted

    @staticmethod
    def hll_concat_userday(record: lbsn.Post) -> str:
        """Merge date and user pkey (guid) for measuring user days"""
        date_merged = HLLFunctions.hll_concat_date(record)
        user_merged = HLLFunctions.hll_concat_user(record)
        pud_hll = HLLFunctions.hll_concat(
            [user_merged, date_merged])
        return pud_hll

    @staticmethod
    def hll_concat_yearmonth(record: lbsn.Post) -> str:
        """Concat year and month of post date to string"""
        date_merged = HLLFunctions.merge_dates_post(record)
        date_formatted = date_merged.strftime('%Y-%m')
        return date_formatted

    @staticmethod
    def hll_concat_latlng(record: lbsn.Post) -> str:
        """Concat post lat lng coordinates to string"""
        if record.post_geoaccuracy == 'latlng':
            coordinates_geom = HF.null_check(
                record.post_latlng)
            coordinates = HF.get_coordinates_from_ewkt(
                coordinates_geom
            )
            return f'{coordinates.lat}:{coordinates.lng}'
        return '0:0'

    @staticmethod
    def hll_concat_place(record: lbsn.Post) -> str:
        """Concat origin and post place guid to string"""
        if record.post_geoaccuracy == 'place':
            return HLLFunctions.hll_concat(
                [record.pkey.origin.origin_id, record.place_pkey.id])
        return None

    @staticmethod
    def hll_concat_upt_hll(record: lbsn.Post) -> List[str]:
        """Concat all post terms (body, title, hashtags) and return list"""
        body_terms = HF.select_terms(
            record.post_body)
        title_terms = HF.select_terms(
            record.post_title)
        tag_terms = {item.lower() for item in record.hashtags if len(item) > 2}
        all_post_terms = set.union(
            body_terms, title_terms, tag_terms)
        user_hll = HLLFunctions.hll_concat_user(
            record)
        upt_hll = HLLFunctions.hll_concat_user_terms(
            user_hll, all_post_terms)
        return upt_hll

    @staticmethod
    def hll_concat_user_terms(user: str, terms: List[str]) -> Set[str]:
        """Concat user and list of terms to set of unique user:terms"""
        concat_user_terms = set()
        for term in terms:
            concat_user_terms.add(":".join([user, term]))
        return concat_user_terms

    @staticmethod
    def hll_concat(record_attr: List[Union[str, int]]) -> str:
        """Concat record attributes using colon """
        return ":".join(map(str, record_attr))

    @staticmethod
    def merge_dates_post(record: lbsn.Post = None) -> Optional[dt.datetime]:
        """Merge post_publish and post_created attributes"""
        post_create_date = HF.null_check_datetime(
            record.post_create_date)
        post_publish_date = HF.null_check_datetime(
            record.post_publish_date)
        if post_create_date is None:
            return post_publish_date
        return post_create_date



# -*- coding: utf-8 -*-

"""
Module for mapping LBSN (RAW) to common LBSN Structure (Protobuf).
"""

# pylint: disable=no-member

import logging
from typing import Optional, Dict, Any
from lbsnstructure import lbsnstructure_pb2 as lbsn
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.duration_pb2 import Duration
from shapely import wkb

from lbsntransform.tools.helper_functions import HelperFunctions as HF

MAPPING_ID = 0

def parse_geom(geom_hex):
    """Parse Postgis hex WKB to geometry WKT"""
    geom = wkb.loads(geom_hex, hex=True)
    return geom.wkt


def set_lbsn_attr(lbsn_obj, attr_name, in_record, geom: Optional[bool] = None):
    """Sets value for attr_name of lbsn_obj if
    attr_value is not None"""
    attr_value = in_record.get(attr_name)
    if attr_value is None:
        return
    if isinstance(attr_value, (list, dict)):
        if len(attr_value) == 0:
            return
        getattr(lbsn_obj, attr_name).extend(attr_value)
        return
    if geom:
        attr_value = parse_geom(attr_value)
    setattr(lbsn_obj, attr_name, attr_value)


def copydate_lbsn_attr(lbsn_obj_attr, copy_from_val):
    """Some protobuf fields cannot be assigned directly,
    this function applies copyfrom assignment"""
    date_pb = Timestamp()
    date_pb.FromDatetime(
        copy_from_val)
    lbsn_obj_attr.CopyFrom(
        date_pb)

def copyduration_lbsn_attr(lbsn_obj_attr, copy_from_val):
    """Some protobuf fields cannot be assigned directly,
    this function applies copyfrom assignment"""
    duration_pb = Duration()
    duration_pb.FromString(
        copy_from_val)
    lbsn_obj_attr.CopyFrom(
        duration_pb)

def set_lbsn_pkey(lbsn_obj_pkey, pkey_obj, pkey_val, origin_val):
    """Sets value for lbsn_obj_pkey of pkey_obj if
    pkey_val is not None"""
    if pkey_val is None:
        return
    pkey_obj = HF.new_lbsn_record_with_id(
        pkey_obj, pkey_val, origin_val)
    lbsn_obj_pkey.CopyFrom(
        pkey_obj.pkey)


class importer():
    """ Provides mapping function from LBSN (raw) endpoints to
        protobuf lbsnstructure
    """
    ORIGIN_NAME = "LBSN"
    ORIGIN_ID = 0

    def __init__(self,
                 disable_reaction_post_referencing=False,
                 geocodes=False,
                 map_full_relations=False,
                 map_reactions=True,
                 ignore_non_geotagged=False,
                 ignore_sources_set=None,
                 min_geoaccuracy=None):
        # We're dealing with LBSN in this class, lets create the OriginID
        # globally
        # this OriginID is required for all CompositeKeys
        origin = lbsn.Origin()
        origin.origin_id = lbsn.Origin.LBSN
        self.origin = origin
        self.null_island = 0
        # this is where all the data will be stored
        # self.lbsn_records = []
        self.log = logging.getLogger('__main__')  # get the main logger object
        self.skipped_count = 0
        self.skipped_low_geoaccuracy = 0
        # self.disableReactionPostReferencing = disableReactionPostReferencing
        # self.mapFullRelations = mapFullRelations
        # self.geocodes = geocodes

    @classmethod
    def get_func_record(cls, record: Dict[str, Any], input_type: Optional[str] = None):
        """Returns mapping function for input_type"""
        FUNC_MAP = {
            lbsn.Origin().DESCRIPTOR.name: cls.extract_origin,
            lbsn.Country().DESCRIPTOR.name: cls.extract_country,
            lbsn.City().DESCRIPTOR.name: cls.extract_city,
            lbsn.Place().DESCRIPTOR.name: cls.extract_place,
            lbsn.UserGroup().DESCRIPTOR.name: cls.extract_usergroup,
            lbsn.User().DESCRIPTOR.name: cls.extract_user,
            lbsn.Post().DESCRIPTOR.name: cls.extract_post,
            lbsn.PostReaction().DESCRIPTOR.name: cls.extract_postreaction,
            lbsn.Event().DESCRIPTOR.name: cls.extract_event,
        }
        func_map = FUNC_MAP.get(input_type)
        # create origin always the same
        origin = lbsn.Origin()
        origin.origin_id = record.get('origin_id')
        return func_map(record, origin)

    def parse_json_record(
            self, record: Dict[str, Any], input_type: Optional[str] = None):
        """Entry point for LBSN data:

        Attributes:
        record: tuple
            0:    A single row from LBSN, stored as dict
            1:    input_type    Type of LBSN record (User, Post, Place etc.)
        """
        record = self.get_func_record(record, input_type)
        # return list of single item
        return [record]

    @classmethod
    def extract_origin(cls, record, origin):
        origin.name = record.get("name")
        return origin

    @classmethod
    def extract_country(cls, record, origin):
        country = HF.new_lbsn_record_with_id(
            lbsn.Country(),
            record.get('country_guid'),
            origin)
        set_lbsn_attr(country, "name", record)
        geom_center = record.get("geom_center")
        if geom_center:
            setattr(country, "geom_center", parse_geom(geom_center))
        geom_area = record.get("geom_area")
        if geom_area:
            setattr(country, "geom_area", parse_geom(geom_area))
        set_lbsn_attr(country, "url", record)
        set_lbsn_attr(country, "name_alternatives", record)
        return country

    @classmethod
    def extract_city(cls, record, origin):
        city = HF.new_lbsn_record_with_id(
            lbsn.City(),
            record.get('city_guid'),
            origin)
        set_lbsn_attr(city, "name", record)
        geom_center = record.get("geom_center")
        if geom_center:
            setattr(city, "geom_center", parse_geom(geom_center))
        geom_area = record.get("geom_area")
        if geom_area:
            setattr(city, "geom_area", parse_geom(geom_area))
        country_guid = record.get('country_guid')
        if country_guid:
            city.country_pkey.CopyFrom(
                HF.new_lbsn_record_with_id(
                    lbsn.Country(), record.get('country_guid'),
                    origin).pkey)
        set_lbsn_attr(city, "url", record)
        set_lbsn_attr(city, "name_alternatives", record)
        set_lbsn_attr(city, "sub_type", record)
        return city

    @classmethod
    def extract_place(cls, record, origin):
        place = HF.new_lbsn_record_with_id(
            lbsn.Place(),
            record.get('place_guid'),
            origin)
        set_lbsn_attr(place, "name", record)
        set_lbsn_attr(place, "post_count", record)
        set_lbsn_attr(place, "url", record)
        geom_center = record.get("geom_center")
        if geom_center:
            setattr(place, "geom_center", parse_geom(geom_center))
        geom_area = record.get("geom_area")
        if geom_area:
            setattr(place, "geom_area", parse_geom(geom_area))
        city_guid = record.get('city_guid')
        if city_guid:
            set_lbsn_pkey(
                place.city_pkey, lbsn.City(), record.get('city_guid'), origin)
        set_lbsn_attr(place, "name_alternatives", record)
        set_lbsn_attr(place, "place_description", record)
        set_lbsn_attr(place, "place_website", record)
        set_lbsn_attr(place, "place_phone", record)
        set_lbsn_attr(place, "address", record)
        set_lbsn_attr(place, "zip_code", record)
        set_lbsn_attr(place, "attributes", record)
        set_lbsn_attr(place, "checkin_count", record)
        set_lbsn_attr(place, "like_count", record)
        set_lbsn_attr(place, "parent_places", record)
        return place

    @classmethod
    def extract_usergroup(cls, record, origin):
        usergroup = HF.new_lbsn_record_with_id(
            lbsn.UserGroup(),
            record.get('usergroup_guid'),
            origin)
        usergroup.usergroup_name = record.get('usergroup_name')
        usergroup.usergroup_description = record.get('usergroup_description')
        usergroup.member_count = record.get('member_count')
        usergroup.usergroup_createdate = record.get('usergroup_createdate')
        usergroup.user_owner = record.get('user_owner')
        user_owner = record.get('user_owner')
        if user_owner:
            usergroup.user_owner_pkey.CopyFrom(
                HF.new_lbsn_record_with_id(
                    lbsn.User(), record.get('user_owner'),
                    origin).pkey)
        return usergroup

    @classmethod
    def extract_user(cls, record, origin):
        user = HF.new_lbsn_record_with_id(
            lbsn.User(),
            record.get('user_guid'),
            origin)
        set_lbsn_attr(user, "user_name", record)
        set_lbsn_attr(user, "user_fullname", record)
        set_lbsn_attr(user, "follows", record)
        set_lbsn_attr(user, "followed", record)
        set_lbsn_attr(user, "biography", record)
        set_lbsn_attr(user, "post_count", record)
        set_lbsn_attr(user, "url", record)
        set_lbsn_attr(user, "is_private", record)
        set_lbsn_attr(user, "is_available", record)

        lang = record.get('user_language')
        if lang:
            ref_user_language = lbsn.Language()
            ref_user_language.language_short = lang
            user.user_language.CopyFrom(ref_user_language)

        set_lbsn_attr(user, "user_location", record)
        user_location_geom = record.get("user_location_geom")
        if user_location_geom:
            setattr(user, "user_location_geom", parse_geom(user_location_geom))
        set_lbsn_attr(user, "liked_count", record)
        active_since = record.get('active_since')
        if active_since:
            copydate_lbsn_attr(
                user.active_since,
                active_since)
        set_lbsn_attr(user, "profile_image_url", record)
        set_lbsn_attr(user, "user_timezone", record)
        set_lbsn_attr(user, "user_utc_offset", record)
        set_lbsn_attr(user, "user_groups_member", record)
        set_lbsn_attr(user, "user_groups_follows", record)
        set_lbsn_attr(user, "group_count", record)
        return user

    @classmethod
    def extract_post(cls, record, origin):
        post = HF.new_lbsn_record_with_id(
            lbsn.Post(),
            record.get('post_guid'),
            origin)
        post_latlng = record.get("post_latlng")
        if post_latlng:
            setattr(post, "post_latlng", parse_geom(post_latlng))
        place_guid = record.get('place_guid')
        if place_guid:
            set_lbsn_pkey(
                post.place_pkey, lbsn.Place(),
                record.get('place_guid'), origin)
        city_guid = record.get('city_guid')
        if city_guid:
            set_lbsn_pkey(
                post.city_pkey, lbsn.City(),
                record.get('city_guid'), origin)
        country_guid = record.get('country_guid')
        if country_guid:
            set_lbsn_pkey(
                post.country_pkey, lbsn.Country(),
                record.get('country_guid'), origin)
        set_lbsn_pkey(
            post.user_pkey, lbsn.User(),
            record.get('user_guid'), origin)
        pub_date = record.get('post_publish_date')
        if pub_date:
            copydate_lbsn_attr(
                post.post_publish_date,
                pub_date)
        set_lbsn_attr(post, "post_body", record)
        post.post_geoaccuracy
        geo_acc = record.get("post_geoaccuracy")
        if geo_acc:
            # get enum value
            post.post_geoaccuracy = lbsn.Post.PostGeoaccuracy.Value(
                geo_acc.upper())
        set_lbsn_attr(post, "hashtags", record)
        set_lbsn_attr(post, "emoji", record)
        set_lbsn_attr(post, "post_like_count", record)
        set_lbsn_attr(post, "post_comment_count", record)
        set_lbsn_attr(post, "post_views_count", record)
        set_lbsn_attr(post, "post_title", record)
        crt_date = record.get('post_create_date')
        if crt_date:
            copydate_lbsn_attr(
                post.post_create_date,
                crt_date)
        set_lbsn_attr(post, "post_thumbnail_url", record)
        set_lbsn_attr(post, "post_url", record)
        post_type = record.get("post_type")
        if post_type:
            # get enum value
            post.post_type = lbsn.Post.PostType.Value(
                post_type.upper())
        set_lbsn_attr(post, "post_filter", record)
        set_lbsn_attr(post, "post_quote_count", record)
        set_lbsn_attr(post, "post_share_count", record)
        lang = record.get('post_language')
        if lang:
            ref_post_language = lbsn.Language()
            ref_post_language.language_short = lang
            post.post_language.CopyFrom(ref_post_language)
        set_lbsn_attr(post, "input_source", record)
        user_mentions = record.get("user_mentions")
        if user_mentions:
            mentioned_users_list = []
            for user_id in user_mentions:  # iterate over the list
                ref_user_record = \
                    HF.new_lbsn_record_with_id(
                        lbsn.User(), user_id, origin)
                mentioned_users_list.append(ref_user_record)
            post.user_mentions_pkey.extend(
                [user_ref.pkey for user_ref in mentioned_users_list])
        set_lbsn_attr(post, "post_content_license", record)
        return post

    @classmethod
    def extract_event(cls, record, origin):
        event = HF.new_lbsn_record_with_id(
            lbsn.Event(),
            record.get('event_guid'),
            origin)
        set_lbsn_attr(event, "name", record)
        event_latlng = record.get("event_latlng")
        if event_latlng:
            setattr(event, "event_latlng", parse_geom(event_latlng))
        event_area = record.get("event_area")
        if event_area:
            setattr(event, "event_area", parse_geom(event_area))
        set_lbsn_attr(event, "event_website", record)
        event_date = record.get('event_date')
        if event_date:
            copydate_lbsn_attr(
                event.event_date,
                event_date)
        event_date_start = record.get('event_date_start')
        if event_date_start:
            copydate_lbsn_attr(
                event.event_date_start,
                event_date_start)
        event_date_end = record.get('event_date_end')
        if event_date_end:
            copydate_lbsn_attr(
                event.event_date_end,
                event_date_end)
        duration = record.get('duration')
        if duration:
            copyduration_lbsn_attr(
                event.duration,
                duration)
        place_guid = record.get('place_guid')
        if place_guid:
            set_lbsn_pkey(
                event.place_pkey, lbsn.Place(),
                record.get('place_guid'), origin)
        city_guid = record.get('city_guid')
        if city_guid:
            set_lbsn_pkey(
                event.city_pkey, lbsn.City(),
                record.get('city_guid'), origin)
        country_guid = record.get('country_guid')
        if country_guid:
            set_lbsn_pkey(
                event.country_pkey, lbsn.Country(),
                record.get('country_guid'), origin)
        set_lbsn_pkey(
            event.user_pkey, lbsn.User(),
            record.get('user_guid'), origin)
        set_lbsn_attr(event, "event_description", record)
        set_lbsn_attr(event, "event_type", record)
        set_lbsn_attr(event, "event_share_count", record)
        set_lbsn_attr(event, "event_like_count", record)
        set_lbsn_attr(event, "event_comment_count", record)
        set_lbsn_attr(event, "event_views_count", record)
        set_lbsn_attr(event, "event_engage_count", record)
        return event

    @classmethod
    def extract_postreaction(cls, record):
        raise NotImplementedError(
            "Mapping of post reactions is not yet implemented")



# -*- coding: utf-8 -*-

"""
Module for sql insert functions for LBSN (hll) db
"""

from lbsntransform.output.hll import hll_bases as hll
from lbsntransform.tools.helper_functions import HelperFunctions as HF


class HLLSql():
    """Maps LBSN Types to hll SQL Structure
    """

    @staticmethod
    def get_sql_attr_coalesce(attr_key: str, facet: str, base: str) -> str:
        """Get coalesce sql for base attr"""
        if attr_key in ('latlng_geom', 'geom_center'):
            return f'''{attr_key} = COALESCE(
                NULLIF(EXCLUDED.{attr_key},
                '{HF.NULL_GEOM_HEX}'),
                {facet}."{base}".{attr_key},
                '{HF.NULL_GEOM_HEX}')
                '''
        return f'''{attr_key} = COALESCE(
            EXCLUDED.{attr_key}, {facet}."{base}".{attr_key})
            '''

    @staticmethod
    def get_sql_hll_coalesce(hll_key: str, facet: str, base: str) -> str:
        """Get coalesce sql for base hll metrics"""
        return f'''{hll_key} = COALESCE(hll_union(EXCLUDED.{hll_key},
            {facet}."{base}".{hll_key}), hll_empty())
            '''

    @staticmethod
    def concat_sep_lists(len_list_concat: int) -> str:
        """Returns separator if list is not empty"""
        if len_list_concat == 0:
            return ''
        return ',\n'

    @staticmethod
    def hll_insertsql(values_str: str, record_type) -> str:
        """Compile SQL insert for hll upsert (update & insert)"""
        facet = record_type.facet
        base = record_type.base
        base_key = hll.BASE_KEY.get(record_type)
        attr_keys = hll.BASE_ATTRS.get(record_type)
        metric_keys = hll.BASE_METRICS.get(record_type)
        insert_sql = \
            f'''
            INSERT INTO {facet}."{base}" (
                {','.join(base_key + attr_keys + metric_keys)})
            VALUES {values_str}
            ON CONFLICT ({', '.join(base_key)})
            DO UPDATE SET
                {', '.join([
                    HLLSql.get_sql_attr_coalesce(
                        attr_key,
                        facet, base) for attr_key in attr_keys])}
                {HLLSql.concat_sep_lists(len(attr_keys))}
                {', '.join([
                    HLLSql.get_sql_hll_coalesce(
                        metric_key,
                        facet, base) for metric_key in metric_keys])};
            '''
        return insert_sql

    @staticmethod
    def get_hmac_hash_sql() -> str:
        """Returns hmac hash sql function"""

        sql = """/* Produce pseudonymized hash of input id with skey
        * - using skey as seed value
        * - sha256 cryptographic hash function
        * - encode in base64 to reduce length of hash
        * - remove trailing '=' from base64 string
        * - return as text
        */
        CREATE OR REPLACE FUNCTION
        extensions.crypt_hash (id text, skey text)
        RETURNS text
        AS $$
            SELECT
                RTRIM(
                    ENCODE(
                        HMAC(
                            id::bytea,
                            skey::bytea,
                            'sha256'),
                        'base64'),
                    '=')
        $$
        LANGUAGE SQL
        STRICT;
        """
        return sql


# -*- coding: utf-8 -*-

"""Hll bases for topical facet
"""

from typing import Tuple, Union

from lbsnstructure import lbsnstructure_pb2 as lbsn
from lbsntransform.output.hll import hll_bases as hll
from lbsntransform.tools.helper_functions import HelperFunctions as HF

FACET = 'topical'


class TopicalBase(hll.HllBase):
    """Intermediate topical base class that extends HllBase
    for HashtagBase, EmojiBase and TermBase
    """

    def __init__(self):
        super().__init__()
        self.metrics['latlng_hll'] = set()
        self.metrics['place_hll'] = set()
        self.metrics['pud_hll'] = set()


class HashtagBase(TopicalBase):
    """Extends Topical Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='hashtag')

    def __init__(self, hashtag: str = None):
        super().__init__()
        self.key['hashtag'] = None
        if hashtag is None:
            # init empty
            return
        self.key['hashtag'] = hashtag.lower()


class EmojiBase(TopicalBase):
    """Extends Topical Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='emoji')

    def __init__(self, emoji: str = None):
        super().__init__()
        self.key['emoji'] = None
        if emoji is None:
            # init empty
            return
        self.key['emoji'] = emoji


class TermBase(TopicalBase):
    """Extends Topical Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='term')

    def __init__(self, term: str = None):
        super().__init__()
        self.key['term'] = None
        if term is None:
            # init empty
            return
        self.key['term'] = term.lower()


class TopicBase(TopicalBase):
    """Extends Topical Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='topic')

    def __init__(self, topic: Tuple[str] = None):
        super().__init__()
        self.key['topic'] = None
        if topic is None:
            # init empty
            return
        self.key['topic'] = topic


class DomainBase(TopicalBase):
    """Extends Topical Base Class"""
    NAME = hll.HllBaseRef(facet=FACET, base='domain')

    def __init__(self, domain: str = None):
        super().__init__()
        self.key['domain'] = None
        if domain is None:
            # init empty
            return
        self.key['domain'] = domain


class TemplateBase(TopicalBase):
    """Example Base class that extends TopicalBase

    Additional steps:

    * in shared_structure_proto_hlldb.py: extract_hll_bases(),
        define which lbsn.records are mapped to this class
    * if multiple TemplateBase can be extracted from a single lbsn record
        (e.g. such as multiple HashtagBase for a single lbsn.Post),
        update process in hll_bases.py: base_factory()
    * in submit_data.py, add TemplateBase to batched_hll_records so
        they are actually used in the mapping procedure
    * if you need specific metrics/measurements that are noit yet defined,
        update HllMetrics and get_hll_metrics()
        in shared_structure_proto_hlldb.py
    """
    # define the name reference (str-identifier) for this class
    # any class with NAME is automatically registered
    NAME = hll.HllBaseRef(facet=FACET, base='topicexample')

    def __init__(self, record: Union[
            lbsn.Post,
            lbsn.Place] = None):
        """Initialize TemplateBase from lbsn record

        Define here from which lbsn record types
        this base can be initialized, example: lbsn.Post, lbsn.Place
        """
        # initialize parent class (TopicalBase)
        super().__init__()

        # TemplateBase can also be initialized empty,
        # for this reason we specify the structure
        # before mapping record values
        self.key['topic_base_id'] = None
        self.attrs['topic_attr1'] = None
        self.attrs['topic_attr2'] = None
        self.attrs['topic_attr3'] = None
        # TemplateBase inherits from TopicalBase
        # thus, metrics latlng_hll, place_hll and pud_hll
        # are already defined. Specify additional metrics below
        # or remove inheritance and define from scratch
        self.metrics['utl_hll'] = set()
        # if no record is supplied, init empty
        if record is None:
            return
        # now it's time to extract (map) lbsn record to this base:
        # * first define how the key is extracted
        self.key['topic_base_id'] = self.extract_topic_base_id(record)
        # * afterwards define how additional (optional) attributes are
        #    extracted
        self.key['topic_attr1'] = self.extract_topic_base_attrs(record)

    @classmethod
    def extract_topic_base_id(cls, record: Union[lbsn.Post, lbsn.Place]):
        """Template method: define how base key is extracted from record"""
        return

    @classmethod
    def extract_topic_base_attrs(cls, record: Union[lbsn.Post, lbsn.Place]):
        """Template method: define how additional attrs are
        extracted from record
        """
        return


class TermLatLngBase(hll.HllBase):
    """Composite Base (c-base) that extends from HLL base Class

    Note: To distinguish c-bases which are composite bases combining
    aspects from multiple facets, they're termed with a leading underscore
    """
    NAME = hll.HllBaseRef(facet=FACET, base='_term_latlng')

    def __init__(self, record: lbsn.Post = None, term: str = None):
        super().__init__()
        self.key['latitude'] = None
        self.key['longitude'] = None
        self.key['term'] = None
        self.attrs['latlng_geom'] = None
        self.metrics['pud_hll'] = set()
        if term is None:
            # init empty
            return
        self.key['term'] = term.lower()
        if record is None:
            # init empty
            return
        if isinstance(record, lbsn.Post):
            coordinates_geom = record.post_latlng
            coordinates = HF.get_coordinates_from_ewkt(
                coordinates_geom
            )
            self.key['latitude'] = coordinates.lat
            self.key['longitude'] = coordinates.lng
            # additional (optional) attributes
            # formatted ready for sql upsert
            self.attrs['latlng_geom'] = HF.return_ewkb_from_geotext(
                coordinates_geom)
        else:
            raise ValueError(
                "Parsing of LatLngBase only supported "
                "from lbsn.Post")


class HashtagLatLngBase(hll.HllBase):
    """Composite Base (c-base) that extends from HLL base Class

    Note: To distinguish c-bases which are composite bases combining
    aspects from multiple facets, they're termed with a leading underscore
    """
    NAME = hll.HllBaseRef(facet=FACET, base='_hashtag_latlng')

    def __init__(self, record: lbsn.Post = None, hashtag: str = None):
        super().__init__()
        self.key['latitude'] = None
        self.key['longitude'] = None
        self.key['hashtag'] = None
        self.attrs['latlng_geom'] = None
        self.metrics['pud_hll'] = set()
        if hashtag is None:
            # init empty
            return
        self.key['hashtag'] = hashtag.lower()
        if record is None:
            # init empty
            return
        if isinstance(record, lbsn.Post):
            coordinates_geom = record.post_latlng
            coordinates = HF.get_coordinates_from_ewkt(
                coordinates_geom
            )
            self.key['latitude'] = coordinates.lat
            self.key['longitude'] = coordinates.lng
            # additional (optional) attributes
            # formatted ready for sql upsert
            self.attrs['latlng_geom'] = HF.return_ewkb_from_geotext(
                coordinates_geom)
        else:
            raise ValueError(
                "Parsing of LatLngBase only supported "
                "from lbsn.Post")


class EmojiLatLngBase(hll.HllBase):
    """Composite Base (c-base) that extends from HLL base Class

    Note: To distinguish c-bases which are composite bases combining
    aspects from multiple facets, they're termed with a leading underscore
    """
    NAME = hll.HllBaseRef(facet=FACET, base='_emoji_latlng')

    def __init__(self, record: lbsn.Post = None, emoji: str = None):
        super().__init__()
        self.key['latitude'] = None
        self.key['longitude'] = None
        self.key['emoji'] = None
        self.attrs['latlng_geom'] = None
        self.metrics['pud_hll'] = set()
        if emoji is None:
            # init empty
            return
        self.key['emoji'] = emoji
        if record is None:
            # init empty
            return
        if isinstance(record, lbsn.Post):
            coordinates_geom = record.post_latlng
            coordinates = HF.get_coordinates_from_ewkt(
                coordinates_geom
            )
            self.key['latitude'] = coordinates.lat
            self.key['longitude'] = coordinates.lng
            # additional (optional) attributes
            # formatted ready for sql upsert
            self.attrs['latlng_geom'] = HF.return_ewkb_from_geotext(
                coordinates_geom)
        else:
            raise ValueError(
                "Parsing of LatLngBase only supported "
                "from lbsn.Post")
